/*
    FEARLUS model1-0: AbstractSocialLandManager.m
    Copyright (C) 1999-2002, 2005-2006  Macaulay Institute

    This file is part of FEARLUS model1-0, an agent-based model of land use
    change.

    FEARLUS model1-0 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model1-0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*

    FEARLUS model0-8 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-8 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

/* Implementation of the AbstractSocialLandManager class. This
 * provides a rudimentary mechanism for keeping track of approval and
 * disapproval from year to year. Subclasses can override the methods
 * and do different things if required. 
 */

#import "AbstractSocialLandManager.h"
#import "AbstractSocialSubPopulation.h"
#import "Debug.h"
#import <stdio.h>
#import <stdlib.h>

@implementation AbstractSocialLandManager

/* getSubPopClass -> class of compatible subpopulations
 *
 * Return the class of subpopulations compatible with social land managers
 */

+(Class)getSubPopClass {
  return [AbstractSocialSubPopulation class];
}

/* initialiseWithEnvironment:landAllocator:colour:
 *
 * Initialise the list of triggers.
 */

-(void)initialiseWithEnvironment: (Environment *)e
		   landAllocator: (LandAllocator *)la
			  colour: (int)col {
  [super initialiseWithEnvironment: e landAllocator: la colour: col];

  triggers = [(AbstractSocialSubPopulation *)subPop getTriggerList];
				// N.B. This is not this land manager's own
				// personal copy of this list, so no alteration
				// of it is permitted.
  approvers = [List create: [self getZone]];
  disapprovers = [List create: [self getZone]];
  approval = 0.0;
  disapproval = 0.0;
  approve = 0.0;
  disapprove = 0.0;
}

/* acceptDisapproval:from: -> self
 *
 * Accept an amount of disapproval from the specified land manager
 */

-acceptDisapproval: (double)dis
	      from: (AbstractSocialLandManager *)disapprover {
  disapproval += dis;
  [disapprovers addLast: disapprover];
  
  [Debug verbosity: M(showApproval)
	 write: "Land manager %u accepts disapproval %g from manager %u",
	 pin, dis, [disapprover getPIN]];

  return self;
}

/* acceptApproval:from: -> self
 *
 * Accept an amount of approval from the specified land manager
 */

-acceptApproval: (double)app from: (AbstractSocialLandManager *)approver {
  approval += app;
  [approvers addLast: approver];

  [Debug verbosity: M(showApproval)
	 write: "Land manager %u accepts approval %g from manager %u",
	 pin, app, [approver getPIN]];

  return self;
}

/* prepareApprovals
 *
 * Zero the approval and disapproval count.
 */

-(void)prepareApprovals {
  approval = 0.0;
  disapproval = 0.0;
  approve = 0.0;
  disapprove = 0.0;
  [disapprovers removeAll];
  [approvers removeAll];
}

/* determineApprovals
 *
 * Implement the social approval algorithm. This is given as a list of
 * triggers that is subpopulation specific. Each trigger has the
 * actionManager: method that takes the land manager as argument and
 * checks whether the conditions are met for that trigger to cause
 * approval or disapproval of any of its neighbours. 
 */

-(void)determineApprovals {
  [triggers forEach: M(actionManager:) : self];
}

/* getApproval -> approval
 *
 * Return the approval the land manager has received
 */

-(double)getApproval {
  return approval;
}

/* getDisapproval -> disapproval
 *
 * Return the disapproval the land manager has received
 */

-(double)getDisapproval {
  return disapproval;
}

/* getApprove -> approve
 *
 * Return the approval the land manager has given
 */

-(double)getApprove {
  return approve;
}

/* getDisapprove -> disapprove
 *
 * Return the disapproval the land manager has given
 */

-(double)getDisapprove {
  return disapprove;
}

/* getApprovers -> list of approvers
 *
 * Return the list of land managers who have approved of this land manager
 */

-(id <List>)getApprovers {
  return approvers;
}

/* getDisapprovers -> list of disapprovers
 *
 * Return the list of land managers who have disapproved of this land manager
 */

-(id <List>)getDisapprovers {
  return disapprovers;
}

/* incApprove:
 *
 * Increase the amount of approval the land manager has given (called from
 * the Triggers)
 */

-(void)incApprove: (double)amt {
  approve += amt;
}

/* incDisapprove:
 *
 * Increate the amount of disapproval the land manager has given (called from
 * the Triggers)
 */

-(void)incDisapprove: (double)amt {
  disapprove += amt;
}

@end
