/*
    FEARLUS model1-0: BatchSwarm.m
    Copyright (C) 1999-2006  Macaulay Institute

    This file is part of FEARLUS model1-0, an agent-based model of land use
    change.

    FEARLUS model1-0 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model1-0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the BatchSwarm class.

*/

#import "BatchSwarm.h"
#import "ModelSwarm.h"
#import "Environment.h"
#import "LandAllocator.h"
#import "FearlusArguments.h"
#import "Progress.h"
#import "Parameter.h"
#import "LandUse.h"
#import "MiscFunc.h"		// getUsableFileNameFrom:withSuffix:
#import "SubPopulation.h"
#import "Reporter.h"

@implementation BatchSwarm

/* +create
 *
 * Create a BatchSwarm object.
 */

+create: (id)z {
  BatchSwarm *o;

  o = [super create: z];

  return(o);
}

/* -buildReporter
 *
 * Build the reporter object, and its schedule
 */

-buildReporter {
  reporter = [Reporter createBegin: self];
  [reporter setParameters: parameter];
  [reporter setModelSwarm: modelSwarm];
  reporter = [reporter createEnd];
  reporterSchedule = [Schedule createBegin: self];
  [reporterSchedule setRepeatInterval: 1];
  reporterSchedule = [reporterSchedule createEnd];
  [reporterSchedule at: 0 createActionTo: reporter message: M(report)];
  [reporterSchedule activateIn: self];

  return self;
}

/* -buildObjects
 *
 * Build the objects needed for the class
 */

-buildObjects {
  [super buildObjects];
  // Build the parameter object
  parameter = [Parameter createBegin: self];
  if([MiscFunc fileExists: [FearlusArguments getModelFile]]) {
				// Test to see if a model file exists before
                                // loading it.
    [Progress detail: PROGRESS_INIT
	      write: "Loading model parameters from file %s",
	      [FearlusArguments getModelFile]];
    [parameter loadFromFileNamed: [FearlusArguments getModelFile]];
  }
  else {
    if([FearlusArguments modelFileHasBeenSpecified]) {
      fprintf(stderr, "Model parameter file %s not found\n",
	      [FearlusArguments getModelFile]);
    }
    else {
      fprintf(stderr, "No model parameter file specified on the command "
	      "line\n");
    }
    abort();
  }
  [parameter setSpecifiedSeed: seed];
  if([FearlusArguments postInitSeedHasBeenSpecified]) {
    [parameter setPostInitSeed: [FearlusArguments getPostInitSeedArg]];
  }
  parameter = [parameter createEnd];
  [parameter finalise];
  // Build the model swarm -- must wait until parameter finalised.
  modelSwarm = [ModelSwarm createBegin: self];
  [modelSwarm setParameters: parameter];
  modelSwarm = [modelSwarm createEnd];

  [modelSwarm buildObjects];

  // Have to put in these colour management things -- would be more elegant
  // if Environment and LandAllocator knew about being run in the background I
  // know. The 128 and 64 are values that (I hope) won't cause any problems.

  [[modelSwarm getLandAllocator] setMinColour: 64 max: 128];

  return(self);
}


/* -buildActions
 *
 * Override the Swarm buildActions method to introduce the scheduling
 * required for this class.
 */

-buildActions {
  [super buildActions];
  [modelSwarm buildActions];

  initialSchedule = [Schedule createBegin: self];
  initialSchedule = [initialSchedule createEnd];
  if([FearlusArguments useReporter]) {
    [initialSchedule at: 0 createActionTo: self message: M(buildReporter)];
  }
  if(![parameter infiniteTime]) {
    [initialSchedule at: [parameter maximumYear]
		     createActionTo: self
		     message: M(stop)];
  }
  else {
    fprintf(stderr, "%s -- Batch mode and infinite time -- not a good "
	    "combination!\n", sel_get_name(_cmd));
    abort();
  }

  return(self);
}

/* -activateIn
 *
 * Override the Swarm activateIn method to activate the schedules
 */

-activateIn: (id)swarmContext {
  [super activateIn: swarmContext];
  [modelSwarm activateIn: self];	// model schedule must come first
  // (This means it will get done first, so the display always occurs after
  // the model cycle)
  [initialSchedule activateIn: self];
  return([self getSwarmActivity]);
}

/* -go
 *
 * Following the Mousetrap BatchSwarm example, this object implements
 * a go method that is inherited in the ObserverSwarm from the
 * GUISwarm object. We have to implement our own, however.
 */

-go {
  [[self getActivity] run];             // Run it!
 
  return [[self getActivity] getStatus];
}

/* -stop
 *
 * Stop the simulation (again, following the Mousetrap BatchSwarm
 * example).  Call the model swarm to print an ontology.
 */

-stop {
  if([FearlusArguments ontologyFileHasBeenSpecified]) [modelSwarm ontology];

  [getTopLevelActivity() terminate];
 
  return self;
}

/* -setSpecifiedSeed:
 *
 * If the user specified a seed on the command line, then it should be
 * recorded in the parameter output. So it needs to be passed in from
 * main.
 */

-(void)setSpecifiedSeed: (unsigned)value {
  seed = value;
}

/* -drop
 *
 * Override the Swarm drop method to make sure all items created by
 * this object are dropped
 */

-(void)drop {
  [modelSwarm drop];
  [super drop];
}

@end
