/*
    FEARLUS model1-0: CAMELRewardingGovernment.m
    Copyright (C) 1999-2006  Macaulay Institute

    This file is part of FEARLUS model1-0, an agent-based model of land use
    change.

    FEARLUS model1-0 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model1-0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

/* Implementation of RewardingGovernment class. As with all Government
 * classes, class variables are used to store the parameters. 
 */

#import "CAMELRewardingGovernment.h"
#import "Environment.h"
#import "LandAllocator.h"
#import "Parameter.h"
#import "AbstractLandManager.h"
#import "FearlusOutput.h"
#import "Debug.h"
#import "MiscFunc.h"
#import "Grid.h"
#import <stdio.h>
#import <errno.h>
#import <objc/objc-api.h>
#import <unistd.h> // for sleep()

#define MAX_YEARS 10000
 
/* Class variables to store the parameters. These will get duplicated
 * into the instance variables.
 */

static double param_pollutionThreshold = 0.0;
static BOOL given_pollutionThreshold = NO;
static double param_reward = 0.0;
static BOOL given_reward = NO;
static char *param_filenameKey;
static BOOL given_filenameKey = NO;
static unsigned int param_roundsToWait = 0;
static BOOL given_roundsToWait = NO;
static unsigned int param_secondsPerRound = 0;
static BOOL given_secondsPerRound = NO;
static unsigned int param_offsetYear = 0;
static BOOL given_offsetYear = NO;

/* Implementation
 */

@implementation CAMELRewardingGovernment

/* +writeParameters:
 *
 * Write the parameters to a file.
 */

//METHOD TO CORRECT
-(void)writeParameters: (FILE *)fp {
  if(!given_pollutionThreshold || !given_reward || !given_filenameKey || 
     !given_secondsPerRound || !given_roundsToWait || !given_offsetYear) {
    //fprintf(stderr, "Panic: Reward, or Pollution Threshold, or finame key "
    // "has not been specified, but %s has been asked to write its "
    // "parameters\n", class_get_class_name(self));
    abort();
  }
  fprintf(fp, "Pollution threshold\t%g%s", param_pollutionThreshold,
	  [FearlusOutput nl]);
  fprintf(fp, "Low pollution reward per land manager\t%g%s",
	  param_reward, [FearlusOutput nl]);
  fprintf(fp, "Filename key\t%s%s",
	  param_filenameKey, [FearlusOutput nl]);
  fprintf(fp, "Seconds per round\t%u%s",
	  secondsPerRound, [FearlusOutput nl]);
  fprintf(fp, "Rounds to wait for CAMEL\t%u%s",
	  roundsToWait, [FearlusOutput nl]);
  fprintf(fp, "Offset year from CAMEL\t%u%s",
	  offsetYear, [FearlusOutput nl]);
}

+(void)loadParameters: (char *)filename {
  FILE *fp;
  int fscanf_retval;
  long pos;
  int nchar;
  char *dummyStr;

  fp = fopen(filename, "r");
  if(fp == NULL) {
    fprintf(stderr, "Error: Could not open Government file: ");
    perror(filename);
    abort();
  }

  fscanf_retval = fscanf(fp, " PollutionThreshold: %lf",
			 &param_pollutionThreshold);
  if(fscanf_retval == 0 || fscanf_retval == EOF) {
    if(errno != 0) {
      fprintf(stderr, "Error: File read error in Government file: ");
      perror(filename);
      abort();
    }
    else {
      fprintf(stderr, "Invalid format in Government file: %s\n", filename);
      [MiscFunc fileHere: fp];
      abort();
    }
  }

  given_pollutionThreshold = YES;

  fscanf_retval = fscanf(fp, " Reward: %lf", &param_reward);

  if(fscanf_retval == 0 || fscanf_retval == EOF) {
    if(errno != 0) {
      fprintf(stderr, "Error: File read error in Government file: ");
      perror(filename);
      abort();
    }
    else {
      fprintf(stderr, "Invalid format in Government file: %s\n", filename);
      [MiscFunc fileHere: fp];
      abort();
    }
  }

  given_reward = YES;

  // Now let's read the param_filenameKey
  // First, check that the first word is "FilenameKey:"

  fscanf(fp, "%*[ \f\n\r\t\v]");
				// Skip over any whitespace
  pos = ftell(fp);
 
  if(pos == -1) {
    fprintf(stderr, "Failed to set bookmark in file %s: %s\n", filename,
	    strerror(errno));
    abort();
  }

  fscanf(fp, "%*s%n", &nchar);	// Find out the number of characters needed to
  // read in the next word, without reading it in

  if(fseek(fp, pos, SEEK_SET) == -1) {
    fprintf(stderr, "Failed to return to bookmark in file %s: %s\n", filename,
	    strerror(errno));
    abort();
  }

  dummyStr = [globalZone alloc: (nchar + 1) * sizeof(char)];
  dummyStr[nchar] = '\0';	// NULL terminate the string
  fscanf(fp, "%s", dummyStr);		// Read in the word to a buffer big 
                                        // enough to contain it.

  if(strcmp(dummyStr,"FilenameKey:") != 0) {
    fprintf(stderr, "Invalid format in Government file: %s\n", filename);
    [MiscFunc fileHere: fp];
    abort();
  }		
  [globalZone free: dummyStr];

  // Now read the filename
	
  fscanf(fp, "%*[ \f\n\r\t\v]");
				// Skip over any whitespace
  pos = ftell(fp);
 
  if(pos == -1) {
    fprintf(stderr, "Failed to set bookmark in file %s: %s\n", filename,
	    strerror(errno));
    abort();
  }

  fscanf(fp, "%*s%n", &nchar);	// Find out the number of characters needed to
  // read in the next word, without reading it in

  if(fseek(fp, pos, SEEK_SET) == -1) {
    fprintf(stderr, "Failed to return to bookmark in file %s: %s\n", filename,
	    strerror(errno));
    abort();
  }

  param_filenameKey = [globalZone alloc: (nchar + 1) * sizeof(char)];
       // param_filenameKey is used and freed in the configure method.

  param_filenameKey[nchar] = '\0';	// NULL terminate the string

  fscanf(fp, "%s", param_filenameKey);	// Read in the word to a buffer big 
                                        // enough to contain it.

  given_filenameKey = YES;

  fscanf_retval = fscanf(fp, " SecondsPerRound: %u", &param_secondsPerRound);
  if(fscanf_retval == 0 || fscanf_retval == EOF) {
    if(errno != 0) {
      fprintf(stderr, "Error: File read error in Government file: ");
      perror(filename);
      abort();
    }
    else {
      fprintf(stderr, "Invalid format in Government file: %s\n", filename);
      [MiscFunc fileHere: fp];
      abort();
    }
  }

  given_secondsPerRound = YES;

  fscanf_retval = fscanf(fp, " RoundsToWait: %u", &param_roundsToWait);
  if(fscanf_retval == 0 || fscanf_retval == EOF) {
    if(errno != 0) {
      fprintf(stderr, "Error: File read error in Government file: ");
      perror(filename);
      abort();
    }
    else {
      fprintf(stderr, "Invalid format in Government file: %s\n", filename);
      [MiscFunc fileHere: fp];
      abort();
    }
  }

  given_roundsToWait = YES;

  fscanf_retval = fscanf(fp, " OffsetYear: %u", &param_offsetYear);
  if(fscanf_retval == 0 || fscanf_retval == EOF) {
    if(errno != 0) {
      fprintf(stderr, "Error: File read error in Government file: ");
      perror(filename);
      abort();
    }
    else {
      fprintf(stderr, "Invalid format in Government file: %s\n", filename);
      [MiscFunc fileHere: fp];
      abort();
    }
  }

  given_offsetYear = YES;

  fclose(fp);

}

/* configure -> self
 *
 * Configure the RewardingGovernment. This simply means copying the
 * class parameters to the instance variables.
 */

-configure {
  char dummyBuf[3];
  int maxYears = [parameter infiniteTime] ? 
    MAX_YEARS : (int)[parameter maximumYear]; 

  if(!given_pollutionThreshold || !given_reward || !given_filenameKey || 
     !given_secondsPerRound || !given_roundsToWait || !given_offsetYear) {
    fprintf(stderr, "Panic: Attempt to configure %s instance "
	    "when parameters have not been supplied\n",
	    object_get_class_name(self));
    abort();
  }

  pollutionThreshold = param_pollutionThreshold;
  reward = param_reward;
  secondsPerRound = param_secondsPerRound;
  roundsToWait = param_roundsToWait;
  offsetYear = param_offsetYear;

  // The name of the file is 
  // "CAMEL-" + param_filenameKey + "-" + YEAR + ".txt"

  nCharsForYear	= snprintf(dummyBuf, 3, "%d", offsetYear + maxYears);

  filenameToRead = [[self getZone] alloc: 
				     (6 + strlen(param_filenameKey) + 1 +
				      nCharsForYear + 5) * sizeof(char)];
  // filenameToRead will not be freed until the global zone is dropped.
  sprintf(filenameToRead, "CAMEL-%s-", param_filenameKey);

  landUseMapFilename = [[self getZone] alloc: 
					 (8 + strlen(param_filenameKey) + 1 +
					  nCharsForYear + 5) * sizeof(char)];
  // landUseMapFilename will not be freed until the global zone is dropped.
  sprintf(landUseMapFilename, "FEARLUS-%s-", param_filenameKey);

  [globalZone free: param_filenameKey];

  return self;
}

/* govern
 *
 * So far it just waits until the appropriate file has been created.
 */

-(void)govern {
  int i = 0;
  char *tmpStr;

  //[self printLandUseMapToFile];

  // Build the filename of the file to read this particular year.

  tmpStr = [scratchZone alloc: (strlen(filenameToRead) +
				nCharsForYear + 5) * sizeof(char)];
  strcpy(tmpStr, filenameToRead);

  sprintf(tmpStr + strlen(filenameToRead),
	  "%0*d.txt", nCharsForYear, (int) (offsetYear + [env getYear]));

  for(i = 1; i <= roundsToWait; i++) {
    if([MiscFunc fileExists: tmpStr]) break;
    sleep(secondsPerRound);
  }

  if(![MiscFunc fileExists: tmpStr]) {
    fprintf(stderr, "I've waited for file %s to appear for longer than "
	    " %u seconds. Thus, I abort.\n",
	    tmpStr, secondsPerRound * roundsToWait);
    //abort();
  }

  //[self loadPollutionMap];

  [scratchZone free: tmpStr];
}

/* printLandUseMapToFile
 *
 * It prints the land use map in a format suitable for CAMEL.
 */	

-(void)printLandUseMapToFile {
  Grid *myGrid;
  char *tmpStr;

  tmpStr = [scratchZone alloc: (strlen(landUseMapFilename) +
				nCharsForYear + 1) * sizeof(char)];

  strcpy(tmpStr, landUseMapFilename);
  sprintf(tmpStr + strlen(landUseMapFilename),
	  "%0*d.txt", nCharsForYear, (int) (offsetYear + [env getYear]));
  

  myGrid = [env createGrid: tmpStr zone: scratchZone];

  //WARNING!!!

  [myGrid save: "*FEARLUS-LandUseID"];

  [myGrid drop];

  [scratchZone free: tmpStr];

}

-(void)drop {
  [[self getZone] free: filenameToRead];
  [[self getZone] free: landUseMapFilename];
  [super drop];
}


@end;
