/*
    FEARLUS model1-0: CBRCaseBase.m
    Copyright (C) 1999-2006  Macaulay Institute

    This file is part of FEARLUS model1-0, an agent-based model of land use
    change.

    FEARLUS model1-0 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model1-0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/

/* Implementation of case base class. The case base lists for each
 * land use are treated as stacks -- i.e. the most recent is always at
 * the front of the list. 
 */

#import "CBRCaseBase.h"

#import "AssocArray.h"
#import "CBRCase.h"
#import "LandUse.h"
#import "CBRState.h"
#import "Parameter.h"
#import "CBRSimilarity.h"
#import "MiscFunc.h"
#import "FearlusStream.h"

id <Zone> all_cases_zone;
BOOL created_all_cases_zone = NO;

@implementation CBRCaseBase

/* create:parameters: -> new case base
 *
 * Create a new case base. Initialise the associative array with
 * enough space to store each land use in the simulation. This
 * requires the parameter object to be passed in. 
 */

+create: (id <Zone>)z parameters: (Parameter *)parameter {
  CBRCaseBase *obj = [super create: z];

  if(!created_all_cases_zone) {
    all_cases_zone = [Zone create: globalZone];
    created_all_cases_zone = YES;
  }
  obj->case_zone = [Zone create: all_cases_zone];
  obj->case_base = [AssocArray create: obj->case_zone size: [parameter nUses]];

  return obj;
}

/* getAllZones
 *
 * Return the zone in which all case bases are stored.
 */

+(id <Zone>)getAllZones {
  if(!created_all_cases_zone) return nil;
  return all_cases_zone;
}

/* addCase:
 *
 * Add a case to the case base. If the case has a land use that is not
 * present in the case base then create a new list and add it to the
 * associative array keyed with the land use. The case is cloned so it
 * belongs to the case base zone.
 */

-(void)addCase: (CBRCase *)acase {
  LandUse *lu = [acase getDecision];

  if(![case_base keyPresent: lu]) {
    id <List> l = [List create: case_zone];
    if(![case_base addObject: l withKey: lu]) {
      fprintf(stderr, "PANIC: Failed to add case to case base\n");
      abort();
    }
  }
  [(id <List>)[case_base getObjectWithKey: lu]
	      addFirst: [acase clone: case_zone]];
}

/* bestCaseForDecision:state: -> most appropriate case
 *
 * Return the most appropriate case in the case base for the decision
 * and state of the world passed as arguments. This method will return
 * nil if there are no cases for the decision in the case base. A new
 * case object is created as a copy of the case in the case base so
 * that the calling object cannot mess up the case in the case
 * base. It will be the responsibility of that object to drop the case
 * created. 
 */

-(CBRCase *)bestCaseForDecision: (LandUse *)lu state: (CBRState *)state {
  id <List> cases = (id <List>)[case_base getObjectWithKey: lu];
  id <List> considered;
  id <List> rejected;
  CBRCase *best;

  if(cases == nil) return nil;
  if([cases getCount] == 0) return nil;

  considered = [List create: scratchZone];
  rejected = [List create: scratchZone];

  best = (CBRCase *)[cases removeFirst];

  while([cases getCount] > 0) {
    CBRCase *current_case;
    CBRSimilarity *cmp;

    if([[best getState] equalState: state]) break;

    current_case = (CBRCase *)[cases removeFirst];

    cmp = [current_case comparedWith: best relativeTo: state];
    if([cmp more: NO]) {
      [considered addFirst: best];
      best = current_case;
    }
    else if([cmp less: NO]) {
      [rejected addFirst: current_case];
    }
    else {
      [considered addFirst: current_case];
    }
    [cmp drop];
  }
  
  [MiscFunc shuffleList: rejected];
  [MiscFunc shuffleList: considered];

  while([rejected getCount] > 0) {
    [cases addFirst: [rejected removeFirst]];
  }

  while([considered getCount] > 0) {
    [cases addFirst: [considered removeFirst]];
  }

  [rejected drop];
  [considered drop];

  [cases addFirst: best];

  return [best clone: scratchZone];
}

/* printToFile:
 *
 * Print the entire case base to the specified file pointer.
 */

-(void)printToFile: (FILE *)fp {
  id <List> lus = [case_base getKeys];
  id ix, lu;

  for(ix = [lus begin: scratchZone], lu = [ix next];
      [ix getLoc] == Member;
      lu = [ix next]) {
    id <List> cases = (id <List>)[case_base getObjectWithKey: lu];

    [cases forEach: M(printToFile:) : (id)fp];
  }
  [ix drop];
}

/* printStream:
 *
 * Print the entire case base to the specified file pointer.
 */

-(void)printStream: (FearlusStream *)stream {
  id <List> lus = [case_base getKeys];
  id ix, lu;

  [stream write: "Case base %p:\n", self];

  for(ix = [lus begin: scratchZone], lu = [ix next];
      [ix getLoc] == Member;
      lu = [ix next]) {
    id <List> cases = (id <List>)[case_base getObjectWithKey: lu];

    [stream write: "Cases for land use %u (%s):\n",
	    [lu getPIN], [lu getLabel]];
    [cases forEach: M(printStream:) : stream];
  }
  [ix drop];
}

/* print -> self
 *
 * Print the entire case base to stdout
 */

-print {
  [self printToFile: stdout];
  return self;
}

/* drop
 *
 * Destroy the case base. Various bits of memory are allocated for the
 * case base, which would make dropping each thing rather
 * tedious. Instead the approach is to create a special zone for the
 * case base into which all objects belonging to the case base are
 * created. Dropping the case_zone will release all of this memory at
 * once.
 */

-(void)drop {
  [case_zone drop];
  [super drop];
}

@end
