/*
    FEARLUS model1-0: LockInReport.m
    Copyright (C) 1999-2006  Macaulay Institute

    This file is part of FEARLUS model1-0, an agent-based model of land use
    change.

    FEARLUS model1-0 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model1-0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


The implementation for the LandUseReport object

*/

#import "LockInReport.h"
#import "LandUse.h"
#import "Parameter.h"
#import "ModelSwarm.h"
#import "FearlusArguments.h"
#import "FearlusOutput.h"
#import "Environment.h"
#import "LandParcel.h"
#import "Number.h"

@implementation LockInReport

/*

setModelSwarm:andParameters:

Set the model swarm and parameters for this report. These will be used to
obtain the data in the report.

*/

-(void)setModelSwarm: (ModelSwarm *)m andParameters: (Parameter *)p {
  parameter = p;
  model = m;

  lockedIn = NO;
  lockedLandUse = nil;
}

/*

reportForYear:toFile:

Create a report of the number of parcels that use each land use at the
end of the year. The land use counter associative array must first be
initialised to zero for each land use. Then we loop through the land
parcels and add one to the appropriate land use counter according to
the land use of the parcel. Then we loop through the land uses and print
out the totals.

Note that [FearlusOutput nl] is used as the string to print for a
new line. This is in case DOS mode has been specified on the command
line.

*/

-(void)reportForYear: (unsigned)year toFile: (FILE *)fp {
  if(lockedIn) {
    fprintf(fp, "Locked-in\tYear:\t%u\tLand Use:\t%u%s", year,
	    [lockedLandUse getPIN], [FearlusOutput nl]);
  }
  else {
    fprintf(fp, "Locked-out\tYear:\t%u%s", year, [FearlusOutput nl]);
  }
}

/* There's no need for a drop method -- the Reporter object will own all Report
   objects, and stores them in its own Zone. When the zone is deleted, all the
   Reports' allocated memory will go with it, so long as we use [self getZone]
   here when allocating memory. */

/*

reportingForYear:

Return YES if we're reporting this year, according to the years we've been
handed. NO otherwise.

This report is slightly different in that it will in addition only report when
there has been a change in lock-in status.

*/

-(BOOL)reportingForYear: (unsigned)year {
  BOOL reportingYear;
  id inx;
  Number *n;

  /* Decide if we should print this year */
  reportingYear = NO;
  if(([fixedYears getCount] > 0)
     && ([(Number *)[fixedYears atOffset: 0] getUnsigned] == year)) {
    /* This is a fixed year to print. So we'll go ahead, but after we've
       removed the head */
    [[fixedYears removeFirst] drop];
    reportingYear = YES;
  }
  else {
    /* Not a fixed year to print, so check all the repeating years */
    for(inx = [repeatYears begin: scratchZone], [inx next];
	[inx getLoc] == Member;
	[inx next]) {
      n = [inx get];
      if((year % [n getUnsigned]) == 0) {
	/* It's a repeating year to print */
	reportingYear = YES;
	break;
      }
    }
    [inx drop];
  }
  if(reportingYear) reportingYear = [self lockInChange];
  return reportingYear;
}

/*

lockInChange

Return YES if there is a change in the lock-in status, NO otherwise.

*/

-(BOOL)lockInChange {
  LandUse *lu = NULL, *thisLU;
  BOOL thisLockedIn, firstLU;
  id inx;
  
  thisLockedIn = YES;
  firstLU = YES;

  for(inx = [[[model getEnvironment] getLandParcels] begin: [self getZone]],
	[inx next];
      [inx getLoc] == Member;
      [inx next]) {
    thisLU = [(LandParcel *)[inx get] getLandUse];
    if(firstLU) {
      lu = thisLU;
      firstLU = NO;
    }
    else if(lu != thisLU) {
      thisLockedIn = NO;
      break;
    }
  }
  [inx drop];
  if(thisLockedIn != lockedIn) {
    if(thisLockedIn) lockedLandUse = lu;
    lockedIn = thisLockedIn;
    return YES;
  }
  else if(lockedIn && lockedLandUse != lu) {
    lockedLandUse = lu;
    return YES;
  }
  else {
    return NO;
  }
}

@end

