/*
    FEARLUS model1-0: ModelSwarm.m
    Copyright (C) 1999-2006  Macaulay Institute

    This file is part of FEARLUS model1-0, an agent-based model of land use
    change.

    FEARLUS model1-0 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model1-0 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/* Implementation for the ModelSwarm class.
 */

#import "ModelSwarm.h"
#import "Environment.h"
#import "LandAllocator.h"
#import "AbstractGovernment.h"
#import "AbstractSocialLandManager.h"
#import "AbstractSubPopulation.h"
#import "FearlusArguments.h"
#import "Debug.h"
#import "Progress.h"
#import "FearlusStream.h"
#import "Parameter.h"
#import "MiscFunc.h"
#import "ClassInfo.h"
#import "OWLOntology.h"
#import <objc/objc-api.h>

@implementation ModelSwarm

/* +createBegin
 *
 * Create a modelswarm, and give it some initial values. Set up a
 * probe for the internal variables. I wonder why the probes should
 * get set up here (for the class). Surely the probes for the class
 * should only be set up once -- not each time the createBegin method
 * of course. Naturally we'll only be building one ModelSwarm, so such
 * things don't matter, but it's a niggle.
 */

+createBegin: (id)z {
  ModelSwarm *m;
  id <ProbeMap> probeMap;

  m = [super createBegin: z];

  probeMap = [EmptyProbeMap createBegin: z];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  [probeMap addProbe: [probeLibrary getProbeForVariable: "parameter"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "landAllocator"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "environment"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "government"
				    inClass: [self class]]];
  [probeLibrary setProbeMap: probeMap For: [self class]];

  return m;
}

/* -createEnd
 *
 * Initialise the variables.
 */

-createEnd {
  structure_ontology = NO;
  structure_ontology_file = NULL;
  return([super createEnd]);
}

/* -setParameters:
 *
 * Set the parameters of the model
 */

-(void)setParameters: (Parameter *)p {
  parameter = p;
}

/* -getEnvironment
 *
 * Return a pointer to the environment
 */

-(Environment *)getEnvironment {
  return(environment);
}

/* -getLandAllocator
 *
 * Return a pointer to the land allocator
 */

-(LandAllocator *)getLandAllocator {
  return(landAllocator);
}

/* -getGovernment
 *
 * Return a pointer to the government
 */

-(AbstractGovernment *)getGovernment {
  return government;
}

/* -initialSeed
 *
 * Method called from the initialisation schedule to set the seed
 * specified for the run.
 */

-initialSeed {
  printf("Using seed: %u\n", [parameter specifiedSeed]);
  [randomGenerator setStateFromSeed: [parameter specifiedSeed]];
  return self;
}

/* -switchSeed
 *
 * Method called from initialisation schedule if user has specified a
 * separate seed after initialisation.
 */

-switchSeed {
  printf("Using new seed: %u\n", [parameter postInitSeed]);
  [randomGenerator setStateFromSeed: [parameter postInitSeed]];
  return self;
}

/* -ontology
 *
 * Create an ontology at the current time in the schedule consisting of all
 * the objects in the model.
 */

-(void)ontology {
  OWLOntology *ontology;

#ifndef DISABLE_GUI
  if(swarmGUIMode) {
    if(!structure_ontology) {
      if([FearlusArguments ontologyURIHasBeenSpecified]) {
	ontology = [OWLOntology create: scratchZone
				file: [FearlusArguments getOntologyFile]
				uri: [FearlusArguments getOntologyURI]];
      }
      else {
	ontology = [OWLOntology create: scratchZone
				filename: [FearlusArguments getOntologyFile]];
      }
      structure_ontology_file = strdup([ontology getFileName]);
      [ontology metadata];
      [ontology addInstanceClassRecursively: parameter];
      [ontology addInstanceClassRecursively: environment];
      [ontology addInstanceClassRecursively: landAllocator];
      [ontology addInstanceClassRecursively: government];
      [ontology drop];
      structure_ontology = YES;
    }
    ontology = [OWLOntology create: scratchZone
			    prefix: [FearlusArguments getOntologyFile]
			    time: [environment getYear]
			    maxtime: [parameter maximumYear]];
    if([FearlusArguments ontologyURIHasBeenSpecified]) {
      [ontology import: [FearlusArguments getOntologyURI]];
    }
    else {
      char *fileuri;

      fileuri = [scratchZone alloc: ((strlen(structure_ontology_file) + 7)
				     * sizeof(char))];
      sprintf(fileuri, "file:/%s", structure_ontology_file);
      [ontology import: fileuri];
      [scratchZone free: fileuri];
    }
    [ontology metadata];
    [ontology addInstanceRecursively: parameter];
    [ontology addInstanceRecursively: environment];
    [ontology addInstanceRecursively: landAllocator];
    [ontology addInstanceRecursively: government];
    [ontology drop];
  }
  else {
#endif
    if([FearlusArguments ontologyURIHasBeenSpecified]) {
      ontology = [OWLOntology create: scratchZone
			      file: [FearlusArguments getOntologyFile]
			      uri: [FearlusArguments getOntologyURI]];
    }
    else {
      ontology = [OWLOntology create: scratchZone
			      filename: [FearlusArguments getOntologyFile]];
    }
    [ontology metadata];
    [ontology addInstanceClassRecursively: parameter];
    [ontology addInstanceClassRecursively: environment];
    [ontology addInstanceClassRecursively: landAllocator];
    [ontology addInstanceClassRecursively: government];
    [ontology addInstanceRecursively: parameter];
    [ontology addInstanceRecursively: environment];
    [ontology addInstanceRecursively: landAllocator];
    [ontology addInstanceRecursively: government];
    [ontology drop];
#ifndef DISABLE_GUI
  }
#endif
}

/* -buildObjects
 *
 * Override the Swarm class buildObjects method to build my own
 * objects.  Build the environmentm, land allocator, government and
 * land parcels.
 */

-buildObjects {
  [super buildObjects];		// Build the super-class objects

  // Build and initialise the environment.

  environment = [Environment createBegin: self];
  [environment setParameters: parameter];
  environment = [environment createEnd];

  // Build and initialise the land allocator.

  [Progress detail: PROGRESS_INIT
	    write: "Creating land allocator"];
  landAllocator = [LandAllocator createBegin: self];
  [landAllocator setParameters: parameter];
  landAllocator = [landAllocator createEnd];

  // Build and initialise the government.

  [Progress detail: PROGRESS_INIT
	    write: "Creating government"];
  if(![parameter noGovernment]) {
    [[parameter governmentClass] loadParameters: [parameter governmentFile]];
    government = [[parameter governmentClass] create: self];
    [government setParameters: parameter
		environment: environment
		allocator: landAllocator];
    [government configure];
  }
  else {
    government = nil;
  }

  // Build and initialise the land parcels and land uses. This used to be
  // in the schedule for some reason... NOTE THAT NO RANDOM NUMBERS SHOULD
  // BE USED BEFORE THIS POINT!

  [self initialSeed];
  [environment createLandParcels];
  [environment createLandUses];


  // Output the parameters if required

  if([Verbosity showParameters]) {
    [Debug verbosity: M(showParameters)
	   write: "Model being built using the following parameters:"];
    [parameter writeParameters: [[Debug getStream] getFilePointer]];
  }

  return(self);
}

/* -buildActions
 *
 * Override the Swarm class buildActions method to build the schedule
 * for this model. Things happen in the following order:
 *
 * 0.	The environment initiates the next year.
 * 1.	The land managers determine the land use for their land parcel(s)
 * 2.	The environment determines the climate for the year
 * 3.	The environment determines the economy for the year
 * 4+0-8	The environment determines the pollution for the year
 * 5.	The environment tells each land parcel to compute its yield
 * 6+0-8?	The government governs (issues rewards/fines for pollution)
 * 7.	The land managers harvest the yield
 * 8+0-8?	Land Managers determine their social approval of each other if
 *	they are social land managers
 * 9.	The land allocator transfers land parcels between land managers
 * 10.	The land allocator removes from the game any land managers with no land
 * 11+0-8	The land managers learn (adapt their variable psychologicals)
 *
 * 
 */

-buildActions {
  [super buildActions];		// You have to let the super class build its
				// actions
  // Build the action group for the actions to take place cycle by cycle
  modelActions = [ActionGroup create: self];
  // The order in which the actions are created is the order in which they will
  // be executed by the scheduler.
  [modelActions createActionTo: environment message: M(newYear)];
				// Environment initiate a new year.
  [modelActions createActionForEach: [parameter subPopulations]
		message: M(resetRemovals)];
				// Subpopulation housekeeping
  [modelActions createActionForEach: [parameter subPopulations]
		message: M(resetBids)];
 				// Subpopulation housekeeping
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(initialiseYear)];
 				// Land manager housekeeping
  [modelActions createActionForEach: [environment getLandParcels]
		message: M(updateLandManager)];
 				// Get land parcels to update manager
  [modelActions createActionForEach: [parameter subPopulations]
		message: M(resetStrategies)];
 				// Subpopulation housekeeping
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(landUseSelectionAlgorithm)];
				// Land managers allocate land uses
  [modelActions createActionForEach: [environment getLandParcels]
		message: M(updateLandUse)];
                                // Land parcels adopt their new land use
  [modelActions createActionTo: environment message: M(determineClimate)];
				// Environment determine the climate
  [modelActions createActionTo: environment message: M(determineEconomy)];
				// Environment determine the economy
  [modelActions createActionTo: environment message: M(determinePollution)];
				// Environment determine the pollution
  [modelActions createActionTo: environment message: M(calculateYield)];
				// Environment calculate the yield
  [modelActions createActionTo: environment message: M(calculateSuitability)];
				// Environment calculate suitability
  if(![parameter noGovernment]) {
    [modelActions createActionTo: government message: M(govern)];
				// Government responds to pollution
  }
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(harvest)];
                                // Land managers harvest their yield
  if([ClassInfo class: [[parameter subPopClass] getLandManagerClass]
		isSubClassOf: [AbstractSocialLandManager class]]) {
    [modelActions createActionForEach: [landAllocator getLandManagers]
		  message: M(prepareApprovals)];
				// Any preparations needed for approvals
    [modelActions createActionForEach: [landAllocator getLandManagers]
		  message: M(determineApprovals)];
				// Land Managers determine how much
				// they approve or disapprove of their
				// neighbours (if they belong to the
				// appropriate class)
  }
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(learn)];
				// Land managers update variable psychological
				// characteristics
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(sellLandParcels)];
				// Land managers put parcels up for sale
  [modelActions createActionTo: landAllocator message: M(notifyParcelSales)];
 				// Land allocator notifies managers of parcels
 				// for sale
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(bidForParcels)];
				// Land managers bid for parcels
  [modelActions createActionTo: landAllocator message: M(transferLandParcels)];
				// Land allocator transfer land parcels
  [modelActions createActionForEach: [landAllocator getLandManagers]
		message: M(incAge)];
                                // Land Managers increment their age
  [modelActions createActionTo: landAllocator message: M(killLosers)];
				// Land allocator kill off losing land mgrs

  modelSchedule = [Schedule createBegin: self];
  [modelSchedule setRepeatInterval: 1];
 
  modelSchedule = [modelSchedule createEnd];
  [modelSchedule at: 0 createAction: modelActions];
				// Put the action group into the schedule
  // Build the action group for the actions to take place in the initial cycle
  // only

  // Initialise the objects created with buildObjects. The environment
  // will build the land parcels and land uses. The land allocator
  // will build the land managers. At the end of the initialisation
  // schedule, we want the following to have happened:
  //
  // - Initial land uses assigned to land parcels
  // - Initial yields computed for each land parcel
  // - All land managers created
  // - All land parcels assigned to land managers

  initialActions = [ActionGroup create: self];

  [initialActions createActionTo: environment message: M(determineClimate)];
  [initialActions createActionTo: environment message: M(determineEconomy)];
  /* NB Climate and economy are determined before land use in the initial cycle
     only -- is this an acceptable anomaly? */
  [initialActions createActionTo: landAllocator
		  message: M(initialiseWithEnvironment:) : environment];
  [initialActions createActionForEach: [parameter subPopulations]
		  message: M(resetStrategies)];
  [initialActions createActionForEach: [landAllocator getLandManagers]
		  message: M(allocateInitialLandUses)];
  [initialActions createActionTo: environment message: M(determinePollution)];
  [initialActions createActionTo: environment message: M(calculateYield)];
  if(![parameter noGovernment]) {
    [initialActions createActionTo: government message: M(govern)];
				// Government responds to pollution
  }

  // If a separate seed is required after initialisation, then add an action
  // at the end of the initialisation phase to set that seed
  // THIS MUST BE THE LAST ACTION
  if([FearlusArguments postInitSeedHasBeenSpecified]) {
    [initialActions createActionTo: self message: M(switchSeed)];
  }
  //  [initialActions createActionTo: modelSchedule 
  //	  message: M(activateIn:) : self];
  // Build the schedule
  initialSchedule = [Schedule createBegin: self];
  initialSchedule = [initialSchedule createEnd];
  [initialSchedule at: 0 createAction: initialActions];
  [initialSchedule at: 1 createActionTo: modelSchedule
		   message: M(activateIn:) : self];
  return(self);
}

/* -activateIn
 *
 * Activate this model swarm in the required context.
 *
 * 
 */

-activateIn: (id)swarmContext {
  [super activateIn: swarmContext];
  [initialSchedule activateIn: self];
  //  [modelSchedule activateIn: self];
  return([self getSwarmActivity]);
}

/* -drop
 *
 * Drop any object created by this object
 */

-(void)drop {
  [environment drop];
  [landAllocator drop];
  if(structure_ontology_file != NULL) free(structure_ontology_file);
  [super drop];
}

@end
